import math
import itertools
import subprocess
import os
from tempfile import NamedTemporaryFile
from dataclasses import dataclass
from pathlib import Path
import importlib.resources
import lzma
import sys
import hashlib

import platformdirs
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import wasmtime


def interpolate(p1, p2, t, t_start=0, t_end=1):
    """ Interpolate along the line from point p1 to point p2 using t as the interpolation variable.
        t_start and t_end set the bounds of t, t_start at p1, and t_end at p2.
        When both interval ends coincide, clips and returns p1.
    """
    if math.isclose(t_start, t_end):
        return p1
    t_range = t_end - t_start
    t = (t - t_start) / t_range
    x1, y1 = p1
    x2, y2 = p2
    dx, dy = x2 - x1, y2 - y1
    return (x1 + t*dx, y1 + t*dy)


def approx_in_range(value, lower, upper):
    """ Approximate range check """
    if math.isclose(value, lower) or math.isclose(value, upper):
        return True
    return lower <= value <= upper


def edge_cycle(points):
    """ From a list of points return an iterator of all edges assuming they are a closed loop:
        [A B C] -> [AB BC CA]
    """
    return itertools.pairwise(itertools.chain(points, points[:1]))


def polygon_is_clockwise(points):
    (x1, y1, i), *_rest = sorted((x, y, i) for i, (x, y) in enumerate(points))
    x0, y0 = points[(i-1)%len(points)]
    x2, y2 = points[(i+1)%len(points)]
    det = (x0*y1 + x1*y2 + x2*y0) - (x2*y1 + x1*y0 + x0*y2)
    return det < 0


def polygon_center_of_mass(polygon):
    # https://en.wikipedia.org/wiki/Centroid
    total_x, total_y = 0, 0
    area = 0
    for (x1, y1), (x2, y2) in edge_cycle(polygon):
        diff = (x1*y2 - x2*y1)
        total_x += (x1 + x2) * diff
        total_y += (y1 + y2) * diff
        area += diff
    area /= 2
    total_x /= 6*area
    total_y /= 6*area
    return total_x, total_y


class WasmApp:
    def __init__(self, wasm_filename, cachedir="kicoil"):
        module_binary = importlib.resources.read_binary(__package__, wasm_filename)

        module_path_digest = hashlib.sha256(__file__.encode()).hexdigest()
        module_digest = hashlib.sha256(module_binary).hexdigest()
        cache_path = Path(os.getenv("KICOIL_CACHE_DIR", platformdirs.user_cache_dir(cachedir)))
        cache_path.mkdir(parents=True, exist_ok=True)
        cache_filename = (cache_path / f'{wasm_filename}-{module_path_digest[:8]}-{module_digest[:16]}')
        
        self.engine = wasmtime.Engine()

        try:
            with cache_filename.open("rb") as cache_file:
                self.module = wasmtime.Module.deserialize(self.engine, lzma.decompress(cache_file.read()))
        except:
            print("Preparing to run {}. This might take a while...".format(wasm_filename), file=sys.stderr)
            self.module = wasmtime.Module(self.engine, module_binary)
            with cache_filename.open("wb") as cache_file:
                cache_file.write(lzma.compress(self.module.serialize(), preset=0))

    def run(self, stdin='', argv=[]):
        with NamedTemporaryFile('r') as stdout_f, NamedTemporaryFile('w') as stdin_f:
            stdin_f.write(stdin)
            stdin_f.flush()

            wasi_cfg = wasmtime.WasiConfig()
            wasi_cfg.argv = argv
            wasi_cfg.stdin_file = stdin_f.name
            wasi_cfg.stdout_file = stdout_f.name
            wasi_cfg.inherit_stderr()

            linker = wasmtime.Linker(self.engine)
            linker.define_wasi()
            store = wasmtime.Store(self.engine)
            store.set_wasi(wasi_cfg)
            self.app = linker.instantiate(store, self.module)
            linker.define_instance(store, "app", self.app)

            try:
                self.app.exports(store)["_start"](store)
            except wasmtime.ExitTrap as trap:
                if trap.code != 0:
                    raise
            return 0, stdout_f.read()


@dataclass(frozen=True)
class SkeletonNode:
    x: float
    y: float
    time: float

    @property
    def pos(self):
        return self.x, self.y


skeleton_wasm = WasmApp('skeleton.wasm')

def compute_skeleton(exterior):
    points_deduplicated = []
    for p1, p2 in edge_cycle(exterior):
        if p2 != p1:
            points_deduplicated.append(p1)
    input_data = '\n'.join(f'{x} {y}' for x, y in points_deduplicated)
    Path('/tmp/debug.txt').write_text(input_data)
    
    rc, data = skeleton_wasm.run(input_data)

    # Parse output: each line is "x1 y1 x2 y2 t1 t2"
    node_map = {}  # Map (x, y, t) to SkeletonNode
    edges = []

    for line in data.strip().split('\n'):
        if not line:
            continue

        parts = line.split()
        if len(parts) != 6:
            continue

        x1, y1, x2, y2, t1, t2 = map(float, parts)

        n1 = (x1, y1, t1)
        if n1 not in node_map:
            node_map[n1] = SkeletonNode(*n1)
        
        n2 = (x2, y2, t2)
        if n2 not in node_map:
            node_map[n2] = SkeletonNode(*n2)

        edges.append((node_map[n1], node_map[n2]))

    nodes = list(node_map.values())
    return nodes, edges


class Skeletonator:
    def __init__(self, poly):
        self.poly = poly
        self.poly_edges = list(zip(poly, poly[1:] + poly[:1]))
        self.circumference = sum(math.dist(a, b) for a, b in self.poly_edges)
        self.skeleton_nodes, self.skeleton_edges = compute_skeleton(exterior=poly)
        self.arc_map = {}
        self.divergent = set()
        self.radius = max(n.time for n in self.skeleton_nodes)
        self.min_radius = self.radius
        for n1, n2 in self.skeleton_edges:
            if n1 in self.arc_map:
                self.divergent.add(n1)
                self.min_radius = min(n1.time, self.radius)
            self.arc_map[n1] = n2
        coord_map = {}
        for n in self.skeleton_nodes:
            p = (round(n.x, 6), round(n.y, 6))
            coord_map[p] = n
        self.node_map = {}
        for x, y in poly:
            p = (round(x, 6), round(y, 6))
            self.node_map[(x, y)] = coord_map[p]
        self.dump_to_pdf('/tmp/test.pdf')

    def iter_arcs(self, p):
        i = 0
        start = self.node_map[p]
        #print('start', start, start in self.arc_map, start in self.divergent)
        while start in self.arc_map and not start in self.divergent:
            end = self.arc_map[start]
            #print('end', i, end)
            i += 1
            yield start, end
            start = end

    def project_arc(self, p, r):
        t = self.radius - r

        for n0, n1 in self.iter_arcs(p):
            if t < 0 or approx_in_range(t, n0.time, n1.time):
                return (n0, n1), interpolate(n0.pos, n1.pos, t, n0.time, n1.time)
        else:
            raise ValueError(f'{r=:.3f} is out of bounds [0, {self.radius - self.min_radius:.4f}]')

    def calc_circumference(self, r):
        projected = [self.project_arc(p, r)[1] for p in self.poly]
        return sum(math.dist(p1, p2) for p1, p2 in zip(projected, projected[1:] + projected[:1]))

    def project_point(self, t, r=None, r_ref=None):
        t %= 1
        if r is None:
            r = self.radius

        if r_ref is None:
            r_ref = r
        
        t_start = 0
        p_cur = None
        _arcs, points_at_r = self.map_circumference(r_ref)
        circumference_at_r = sum(math.dist(p1, p2) for p1, p2 in edge_cycle(points_at_r))
        for (p1, p2), (p1r, p2r) in zip(self.poly_edges, edge_cycle(points_at_r)):
            edge_frac = math.dist(p1r, p2r) / circumference_at_r
            t_end = t_start + edge_frac
    
            if approx_in_range(t, t_start, t_end):
                p1, p2 = self.project_arc(p1, r)[1], self.project_arc(p2, r)[1]
                p_cur = interpolate(p1, p2, t, t_start, t_end)
                return p_cur

            t_start = t_end

    def map_circumference(self, r):
        points, arcs = [], []
        for p in self.poly:
            arc, pt = self.project_arc(p, r)
            arcs.append(arc)
            points.append(pt)
        return arcs, points
    
    def do_spiral(self, t1, t2, r1=None, r2=None):
        if r1 is None:
            r1 = self.radius
        if r2 is None:
            r2 = self.min_radius

        if t2 < t1:
            t1, t2 = t2, t1
            r1, r2 = r2, r1

        def r_interpolate(t):
            t = max(t1, min(t2, t)) # Clip to start/end of spiral
            f = (t - t1) / (t2 - t1)
            return r1 + (r2 - r1) * f

        for t_start in range(math.floor(t1), math.ceil(t2)):
            t_end = t_start + 1
            r_outer = r_interpolate(t_start)
            r_inner = r_interpolate(t_end)
            r_ref = min(r_inner, r_outer) # Handle outward spirals where the radii are swapped
            _ic_arcs, inner_circumference = self.map_circumference(r_ref)

            angle = t_start
            circumference_angles = []
            inner_circumference_sum = sum(math.dist(p1, p2) for p1, p2 in edge_cycle(inner_circumference))
            point_angles = []
            for p1, p2 in edge_cycle(inner_circumference):
                edge_angle = math.dist(p1, p2) / inner_circumference_sum
                point_angles.append(angle)
                angle += edge_angle
            point_angles.append(t_end)

            for (p1, p2), (tp1, tp2) in zip(self.poly_edges, itertools.pairwise(point_angles)):
                rp1 = r_interpolate(tp1)
                rp2 = r_interpolate(tp2)
                _arc, p1_proj = self.project_arc(p1, rp1)
                _arc, p2_proj = self.project_arc(p2, rp2)
                
                if approx_in_range(t1, tp1, tp2):
                    _arc, p2_proj_r1 = self.project_arc(p2, r1)
                    yield interpolate(p1_proj, p2_proj_r1, t1, tp1, tp2), r_ref
                if approx_in_range(t2, tp1, tp2):
                    _arc, p1_proj_r2 = self.project_arc(p1, r2)
                    yield interpolate(p1_proj_r2, p2_proj, t2, tp1, tp2), r_ref
                elif approx_in_range(tp2, t1, t2):
                    yield p2_proj, r_ref

    def dump_to_pdf(self, filename):
        with PdfPages(filename) as pdf:
            fig, ax = plt.subplots(figsize=(10, 10))

            # polygon outline
            poly_x = [p[0] for p in self.poly] + [self.poly[0][0]]
            poly_y = [p[1] for p in self.poly] + [self.poly[0][1]]
            ax.plot(poly_x, poly_y, 'b-', linewidth=2, label='Polygon')
            ax.plot(poly_x, poly_y, 'bo', markersize=4)

            # skeleton edges
            for node1, node2 in self.skeleton_edges:
                ax.plot([node1.x, node2.x], [node1.y, node2.y], 'r-', linewidth=1, alpha=0.7)

            # skeleton nodes
            for n in self.skeleton_nodes:
                if n in self.divergent:
                    ax.plot(n.x, n.y, 'go', markersize=6)
                elif n in self.arc_map:
                    ax.plot(n.x, n.y, 'ro', markersize=3, alpha=0.5)
                else:
                    ax.plot(n.x, n.y, 'o', color='magenta', markersize=6)

            ax.set_aspect('equal', adjustable='box')
            ax.grid(True, alpha=0.3)
            ax.legend()
            ax.set_title(f'Polygon Skeleton (radius: {self.radius:.3f}, min_radius: {self.min_radius:.3f})')
            ax.set_xlabel('X')
            ax.set_ylabel('Y')
            ax.invert_yaxis()
            pdf.savefig(fig, bbox_inches='tight')
            plt.close(fig)



from gerbonara.cad.kicad.pcb import Board, TrackSegment, Via
from gerbonara.cad.kicad.footprints import Atom, AtPos, XYCoord, Pad, Line, Arc, Stroke, Drill

def make_pad(num, layer, x, y, diameter, clearance):
    return Pad(
            number=str(num),
            type=Atom.smd,
            shape=Atom.circle,
            at=AtPos(x=x, y=y),
            size=XYCoord(x=diameter, y=diameter),
            layers=layer,
            clearance=clearance,
            zone_connect=0)

def make_line(x1, y1, x2, y2, trace_width, layer):
    return Line(
                start=XYCoord(x=x1, y=y1),
                end=XYCoord(x=x2, y=y2),
                layer=layer, 
                stroke=Stroke(width=trace_width))

def make_arc(x1, y1, x2, y2, xm, ym, trace_width, layer):
    return Arc(
                start=XYCoord(x=x1, y=y1),
                mid=XYCoord(x=xm, y=ym),
                end=XYCoord(x=x2, y=y2),
                layer=layer, 
                stroke=Stroke(width=trace_width))


def make_via(x, y, diameter, drill, clearance, layers):
    return Pad(number="NC",
                     type=Atom.thru_hole,
                     shape=Atom.circle,
                     at=AtPos(x=x, y=y),
                     size=XYCoord(x=diameter, y=diameter),
                     drill=Drill(diameter=drill),
                     layers=layers,
                     clearance=clearance, 
                     zone_connect=0)


def footprint_to_board(footprint):
    return Board.empty_board(
            zones=zones,
            track_segments=[TrackSegment.from_footprint_line(line) for line in lines],
            vias=[Via.from_pad(pad) for pad in pads if pad.type == Atom.thru_hole])





# Copyright 2025 Jan Sebastian Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import warnings
import logging
from dataclasses import dataclass, field
import sys
from math import *

from gerbonara.cad.kicad.footprints import Footprint
from gerbonara.cad.kicad.primitives import Zone, Hatch, ZoneKeepout, ZonePolygon, XYCoord

from . import kicad, skeletonator


mu_0 = 1.25663706127e-06 # from scipy.constants


def point_line_distance(p, l1, l2):
    x0, y0 = p
    x1, y1 = l1
    x2, y2 = l2
    # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
    return ((x2-x1)*(y1-y0) - (x1-x0)*(y2-y1)) / sqrt((x2-x1)**2 + (y2-y1)**2)


# https://en.wikipedia.org/wiki/Farey_sequence#Next_term
def farey_sequence(n: int, descending: bool = False) -> None:
    """Print the n'th Farey sequence. Allow for either ascending or descending."""
    a, b, c, d = 0, 1, 1, n
    if descending:
        a, c = 1, n - 1
    yield a, b

    while c <= n and not descending or a > 0 and descending:
        k = (n + b) // d
        a, b, c, d = c, d, k * c - a, k * d - b
        yield a, b


def divisors(n, max_b=10):
    for a, b in farey_sequence(n):
        if a != 0 and a == n and b < max_b:
            yield b
        if b != 1 and b == n and a < max_b:
            yield a


def arc_approximate(points, trace_width, layer, tolerance=0.02, level=0):
    """ Approximate spiral arm using circular arcs. This results in a smoother output using less segments than if we
    approximate the arc using straight line segments.

    The input to this function is a list of points of a straight line segment approximation, and it returns a list of
    gerbonara arc objects approximating the input. """ 
    indent = '    ' * level
    if len(points) < 3:
        raise ValueError('Need at least three points to approximate')

    i_mid = len(points)//2

    x0, y0 = points[0]
    x1, y1 = points[i_mid]
    x2, y2 = points[-1]

    if len(points) < 5:
        yield kicad.make_arc(x0, y0, x2, y2, x1, y1, trace_width, layer)

    # https://stackoverflow.com/questions/56224824/how-do-i-find-the-circumcenter-of-the-triangle-using-python-without-external-lib
    d = 2 * (x0 * (y2 - y1) + x2 * (y1 - y0) + x1 * (y0 - y2))
    cx = ((x0 * x0 + y0 * y0) * (y2 - y1) + (x2 * x2 + y2 * y2) * (y1 - y0) + (x1 * x1 + y1 * y1) * (y0 - y2)) / d
    cy = ((x0 * x0 + y0 * y0) * (x1 - x2) + (x2 * x2 + y2 * y2) * (x0 - x1) + (x1 * x1 + y1 * y1) * (x2 - x0)) / d
    r = dist((cx, cy), (x1, y1))
    if any(abs(dist((px, py), (cx, cy)) - r) > tolerance for px, py in points):
        yield from arc_approximate(points[:i_mid+1], trace_width, layer, tolerance, level+1)
        yield from arc_approximate(points[i_mid:], trace_width, layer, tolerance, level+1)

    else:
        if point_line_distance((x1, y1), (x0, y0), (x2, y2)) > 0:
            yield kicad.make_arc(x0, y0, x2, y2, x1, y1, trace_width, layer)
        else:
            yield kicad.make_arc(x2, y2, x0, y0, x1, y1, trace_width, layer)


class Shape:
    pass


@dataclass
class CircleShape(Shape):
    outer_diameter: float
    inner_diameter: float


    def __post_init__(self):
        self.outer_radius = self.outer_diameter / 2
        self.inner_radius = self.inner_diameter / 2


    @property
    def slug(self):
        return 'circle_{self.outer_diameter:.2f}x{self.inner_diameter:.2f}'


    @property
    def desc(self):
        return f'{self.outer_diameter:.2f} mm OD, {self.inner_diameter:.2f} mm ID circular'


    def compute_spiral(self, a1, a2, fn=64):
        r1, r2 = self.outer_radius, self.inner_radius
        fn = ceil(fn * abs(a2-a1)/(2*pi))
        x0, y0 = cos(a1)*r1, sin(a1)*r1
        dr = 3 if r2 < r1 else -3

        xn, yn = x0, y0
        points = [(x0, y0)]
        dists = []
        for i in range(fn):
            xp, yp = xn, yn
            r = r1 + (i+1)*(r2-r1)/fn
            a = a1 + (i+1)*(a2-a1)/fn
            xn, yn = cos(a)*r, sin(a)*r
            points.append((xn, yn))
            dists.append(dist((xp, yp), (xn, yn)))

        return points, sum(dists), [None]*len(points)


    def project_point(self, r, a, r_ref=None):
        return cos(a) * r, sin(a) * r


    def offset_exterior(self, margin):
        r = self.outer_radius + margin
        tol = 0.05 # mm
        n = ceil(pi / acos(1 - tol/r))
        return [(r*cos(a*2*pi/n), r*sin(a*2*pi/n)) for a in range(n)]


class OffsetShape(Shape):
    def __post_init__(self):
        self.sk = skeletonator.Skeletonator(self.polygon)
        self.outer_radius = self.sk.radius
        self.inner_radius = self.sk.radius - self.annular_width

    @property
    def slug(self):
        return 'polygonal_{len(self.polygon)}pt_r{self.radius:.2f}mm'


    @property
    def desc(self):
        return f'polygonal (n={len(self.polygon)} point, r={self.radius:.2f} mm radius)'


    def compute_spiral(self, a1, a2, fn=None):
        # Skeletonator uses a t coordinate from 0 - 1 per revolution instead of a radian angle.
        points = []
        angle_refs = []
        for point, angle_ref in self.sk.do_spiral(a1/(2*pi), a2/(2*pi), self.outer_radius, self.inner_radius):
            points.append(point)
            angle_refs.append(angle_ref)
        if a2 < a1:
            points = points[::-1]
            angle_refs = angle_refs[::-1]
        arm_length = sum(dist(p1, p2) for p1, p2 in zip(points, points[1:]))
        return points, arm_length, angle_refs


    def project_point(self, r, a, r_ref=None):
        # Skeletonator uses a t coordinate from 0 - 1 per revolution instead of a radian angle.
        return self.sk.project_point(a/(2*pi) % 1, r, r_ref=r_ref)


    def offset_exterior(self, margin):
        return self.sk.do_spiral(0, 1, self.outer_radius + margin, self.outer_radius + margin)


@dataclass
class TrapezoidShape(OffsetShape):
    width: float
    height: float
    offset: float
    annular_width: float
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        w, h, d = self.width, self.height, self.offset
        self.polygon = [(w/2-d, -h/2), (w/2, h/2), (-w/2, h/2), (-w/2+d, -h/2)]
        super().__post_init__()
    
    @property
    def slug(self):
        return f'trapezoid_{self.width:.2f}-2*{self.offset:.2f}x{self.height:.2f}'
    
    @property
    def desc(self):
        return f'{self.width:.2f} x {self.height:.2f} mm, {self.offset:.2f} mm offset isosceles trapezoidal'


@dataclass
class RectangleShape(OffsetShape):
    width: float
    height: float
    annular_width: float
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        w, h = self.width, self.height
        self.polygon = [(w/2, -h/2), (w/2, h/2), (-w/2, h/2), (-w/2, -h/2)]
        super().__post_init__()
    
    @property
    def slug(self):
        return f'rectangle_{self.width:.2f}x{self.height:.2f}'
    
    @property
    def desc(self):
        return f'{self.width:.2f} x {self.height:.2f} mm rectangle'

@dataclass
class SectorShape(OffsetShape):
    inner_diameter: float
    outer_diameter: float
    angle: float
    annular_width: float
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        # Careful: The inner/outer radius properties are relative to the polygon center and are very different from these!
        r1, r2 = self.inner_diameter / 2, self.outer_diameter/2
        n1 = ceil(pi / acos(1 - self.arc_tolerance/self.inner_diameter) * self.angle / (2*pi))
        n2 = ceil(pi / acos(1 - self.arc_tolerance/self.outer_diameter) * self.angle / (2*pi))
        # center on y axis
        pt = lambda r, a: (r*sin(a), r*cos(a))
        self.polygon  = [pt(r2, self.angle/2 - i/n1 * self.angle) for i in range(n1+1)]
        self.polygon += [pt(r1, i/n1 * self.angle - self.angle/2) for i in range(n1+1)]
        super().__post_init__()
    
    @property
    def slug(self):
        return f'sector_{self.outer_diameter:.2f}x{self.inner_diameter:.2f}_{degrees(self.angle):.0f}deg'
    
    @property
    def desc(self):
        return f'{self.outer_diameter:.2f} x {self.inner_diameter:.2f} mm {degrees(self.angle):.0f} deg sector'


@dataclass
class StarShape(OffsetShape):
    inner_diameter: float
    outer_diameter: float
    annular_width: float
    points: int = 5
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        # center on y axis
        pt = lambda r, a: (-r*sin(a), r*cos(a))
        circle = lambda r, n, phase: [pt(r, (i + phase)*2*pi/n) for i in range(n)]
        self.polygon  = [x for pair in zip(circle(self.outer_diameter/2, self.points, 0), circle(self.inner_diameter/2, self.points, 0.5)) for x in pair]
        super().__post_init__()
    
    @property
    def slug(self):
        return f'star_{self.outer_diameter:.2f}x{self.inner_diameter:.2f}'
    
    @property
    def desc(self):
        purpose = ', for demonic purposes' if self.points == 5 else ''
        return f'{self.outer_diameter:.2f} x {self.inner_diameter:.2f} mm star shape{purpose}'


@dataclass
class RegularPolygonShape(OffsetShape):
    diameter: float
    annular_width: float
    corners: int = 8
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        # center on y axis
        pt = lambda r, a: (-r*sin(a), r*cos(a))
        circle = lambda r, n, phase: [pt(r, (i + phase)*2*pi/n) for i in range(n)]
        self.polygon  = list(circle(self.diameter/2, self.corners, 0))
        print(self.polygon)
        super().__post_init__()
    
    @property
    def slug(self):
        return f'regular_{self.corners}gon_{self.diameter:.2f}'
    
    @property
    def desc(self):
        return f'{self.diameter:.2f} mm diameter {self.corners} corner regular polygon'


@dataclass
class SVGShape(OffsetShape):
    filename: str
    annular_width: float
    arc_tolerance: float = 0.05 # mm
    polygon: list = field(init=False)

    def __post_init__(self):
        # center on y axis
        from bs4 import BeautifulSoup
        with open(self.filename) as f:
            soup = BeautifulSoup(f.read(), features='xml')
            path = soup.find('path', recursive=True)
            d = path.attrs['d']
            d = d.strip('MmZ ').replace(',', 'L')
            coord_pairs = d.split('L')
            coords = list(reversed([tuple(map(float, pair.split())) for pair in coord_pairs]))
            # Calculate bounding box
            min_x = min(x for x, _y in coords)
            min_y = max(x for x, _y in coords)
            max_x = min(y for _x, y in coords)
            max_y = max(y for _x, y in coords)
            if max_x < 0 or max_y < 0 or min_x > 0 or min_y > 0:
                # (0, 0) is not within the polygon's axis-aligned bounding box, recenter.
                ox, oy = skeletonator.polygon_center_of_mass(coords)
                warnings.warn(f'Polygon looks not centered, bounds are ({min_x:.2f}, {min_y:.2f}), ({max_x:.2f}, {max_y:.2f}). Aligning (0, 0) with polygon centroid at ({ox:.2f}, {oy:.2f})')
                coords = [(x-ox, y-oy) for x, y in coords]
        self.polygon  = coords
        super().__post_init__()
    
    @property
    def slug(self):
        return f'svg_{len(self.polygon)}n'
    
    @property
    def desc(self):
        return f'{len(self.polygon)} node imported SVG shape'

@dataclass
class PlanarInductor():
    shape: Shape
    turns: int
    twists: int
    trace_width: float = None
    clearance: float = None
    layers: int = 2
    via_diameter: float = 0.6
    via_drill: float = None
    via_offset: float = None
    stagger_inner_vias: bool = False
    stagger_outer_vias: bool = False
    keepout_zone: bool = True
    keepout_margin: float = 0.0
    copper_thickness: float = 0.035
    layer_pair: str = 'F.Cu,B.Cu'
    clockwise: bool = False
    approximate_arcs: bool = True

    def __post_init__(self):
        self.logger = logging.getLogger('kicoil')
        self.turns_per_layer = self.turns/self.layers
        self.sector_angle = 2*pi / self.twists

        if self.clockwise:
            self.sector_angle *= -1
            
        self.sweeping_angle = self.sector_angle * self.turns_per_layer
        self.spiral_pitch = (self.shape.outer_radius-self.shape.inner_radius) / self.turns_per_layer
        self.R = None # will be calculated during render

        c1 = self.shape.inner_radius
        c2 = self.shape.inner_radius + self.spiral_pitch
        alpha1 = atan((self.shape.outer_radius - self.shape.inner_radius) / self.sweeping_angle / c1)
        alpha2 = atan((self.shape.outer_radius - self.shape.inner_radius) / self.sweeping_angle / c2)
        alpha = (alpha1+alpha2)/2
        self.projected_spiral_pitch = self.spiral_pitch*cos(alpha)

        if self.layers == 1 and self.twists > 1:
            warnings.warn('Warning: Twists set to a value other than 1, but single-layer mode is enabled. The twists value will be ignored.')
            self.twists = 1

        if self.turns < 1:
            raise ValueError(f'Error: PlanarInductor.turns must be 1 or more')

        if self.twists < 0:
            raise ValueError(f'Error: PlanarInductor.turns must be 0 or more')

        if gcd(self.twists, self.turns) != 1:
            raise ValueError(f'For the geometry to work out, the twists parameter must be co-prime to turns, i.e. the two must have 1 as their greatest common divisor. You can print valid values for twists by running the kicoil CLI with --show-twists [turns number].\n\n'
                                       f'Right now, both are divisible by {gcd(self.twists, self.turns)}.\n'
                                       f'Valid twist counts for n={self.turns} turns are: {list(divisors(self.turns, max(self.turns, 25)))}\n'
                                       f'Valid turn counts for k={self.twists} twists are: {list(divisors(self.twists, max(self.twists, 25)))}')

        if (self.stagger_inner_vias or self.stagger_outer_vias) and self.twists%2 != 0:
            raise ValueError('For via staggering to work, twists must be even and turns must be odd.')

        if self.trace_width is None and self.clearance is None:
            self.clearance = 0.15
            warnings.warn(f'Warning: Neither trace width nor clearance given. Defaulting to {self.clearance:.2f} mm clearance.')

        if self.trace_width is None:
            if round(self.clearance, 3) > round(self.projected_spiral_pitch, 3):
                warnings.warn(f'Error: Given clearance of {clearance:.2f} mm is larger than the projected spiral pitch of {projected_spiral_pitch:.2f} mm. Reduce clearance or increase the size of the coil.')
            self.trace_width = self.projected_spiral_pitch - self.clearance
            self.logger.info(f'Calculated trace width for {self.clearance:.2f} mm clearance is {self.trace_width:.2f} mm.')

        elif self.clearance is None:
            if round(self.trace_width, 2) > round(self.projected_spiral_pitch, 2):
                warnings.warn(f'Error: Given trace width of {self.trace_width:.2f} mm is larger than the projected spiral pitch of {self.projected_spiral_pitch:.2f} mm. Reduce clearance or increase the size of the coil.')
            self.clearance = self.projected_spiral_pitch - self.trace_width
            self.logger.info(f'Calculated clearance for {self.trace_width:.2f} mm trace width is {self.clearance:.2f} mm.')

        else:
            if round(self.trace_width, 2) > round(self.projected_spiral_pitch, 2):
                raise click.ClickException(f'Error: Given trace width of {self.trace_width:.2f} mm is larger than the projected spiral pitch of {self.projected_spiral_pitch:.2f} mm. Reduce clearance or increase the size of the coil.')
            clearance_actual = self.projected_spiral_pitch - self.trace_width
            if round(clearance_actual, 3) < round(self.clearance, 3):
                raise click.ClickException(f'Error: Actual clearance for {self.trace_width:.2f} mm trace is {clearance_actual:.2f} mm, which is lower than the given clearance of {self.clearance:.2f} mm.')

        if round(self.via_diameter, 2) < round(self.trace_width, 2):
            self.logger.warning(f'Clipping via diameter from {self.via_diameter:.2f} mm to trace width of {self.trace_width:.2f} mm.')
            self.via_diameter = self.trace_width

        if self.via_drill is None:
            self.via_drill = max(self.via_diameter / 2, self.via_diameter - 1.2)
            self.logger.warning(f'No via drill given, defaulting to {self.via_drill:.2f} mm based on via diameter.')

        if self.via_offset is None:
            self.via_offset = max(0, (self.via_diameter - self.trace_width)/2)
            self.logger.info(f'Autocalculated via offset {self.via_offset:.2f} mm')

        if isclose(self.via_offset, 0, abs_tol=1e-6):
            self.via_offset = 0

        self.inner_via_ring_radius = self.shape.inner_radius - self.via_offset
        inner_via_angle = 2*asin((self.via_diameter + self.clearance)/2 / self.inner_via_ring_radius)

        self.outer_via_ring_radius = self.shape.outer_radius + self.via_offset
        outer_via_angle = 2*asin((self.via_diameter + self.clearance)/2 / self.outer_via_ring_radius)

        self.logger.info(f'Inner via ring @r={self.inner_via_ring_radius:.2f} mm (from {self.shape.inner_radius:.2f} mm)')
        self.logger.info(f'    {degrees(inner_via_angle):.1f} deg / via')
        self.logger.info(f'Outer via ring @r={self.outer_via_ring_radius:.2f} mm (from {self.shape.outer_radius:.2f} mm)')
        self.logger.info(f'    {degrees(outer_via_angle):.1f} deg / via')

        # Check if the vias of the inner ring are so large that they would overlap
        if inner_via_angle*self.twists > (4*pi if self.stagger_inner_vias else 2*pi):
            min_dia = 2*((self.via_diameter + self.clearance) / (2*sin(pi / self.twists * (2 if self.stagger_inner_vias else 1))) + self.via_offset)
            warnings.warn(f'Overlapping vias in inner via ring. Calculated minimum inner diameter is {min_dia:.2f} mm.')

        t, _, b = self.layer_pair.partition(',')
        self.layer_pair = (t.strip(), b.strip())

        # For fill factor & inductance formulas, See https://coil32.net/pcb-coil.html for details
        d_avg = (2*self.shape.outer_radius + self.shape.inner_radius)/2
        phi = (2*self.shape.outer_radius - 2*self.shape.inner_radius) / (2*self.shape.outer_radius + 2*self.shape.inner_radius)
        c1, c2, c3, c4 = 1.00, 2.46, 0.00, 0.20 # FIXME for other shapes
        self.L = mu_0 * self.turns**2 * d_avg*1e3 * c1 / 2 * (log(c2/phi) + c3*phi + c4*phi**2)
        self.logger.info(f'Outer diameter: {2*self.shape.outer_radius:g} mm')
        self.logger.info(f'Average diameter: {d_avg:g} mm')
        self.logger.info(f'Inner diameter: {2*self.shape.inner_radius:g} mm')
        self.logger.info(f'Fill factor: {phi:g}')
        self.logger.info(f'Approximate inductance: {self.L:g} µH')

        _points, arm_length, _angle_refs = self.shape.compute_spiral(a1=0, a2=self.sector_angle)

        self.track_length = arm_length*self.twists*self.layers
        self.logger.info(f'Approximate track length: {self.track_length:.2f} mm')

        A = self.copper_thickness/1e3 * self.trace_width/1e3 # trace cross-section area
        rho = 1.68e-8 # specific resistivity of copper
        self.R = self.track_length/1e3 * rho / A
        self.logger.info(f'Approximate resistance: {self.R:g} Ω')

    @property
    def default_footprint_name(self):
        return f'planar-coil-{self.shape.slug}-n{self.turns}-k{self.twists}'

    def render_footprint(self, name=None, arc_tolerance=0.02, circle_segments=64):
        if name is None:
            name = self.default_footprint_name

        from . import __version__
        footprint = Footprint(
                name=name,
                generator=kicad.Atom('kicoil'),
                generator_version=__version__,
                layer='F.Cu',
                descr=f"{self.turns} turn {self.shape.desc} twisted coil footprint, inductance approximately {self.L:.6f} µH. Generated by kicoil, version {__version__}.",
                clearance=self.clearance,
                zone_connect=0)

        total_angle = self.twists*2*self.sweeping_angle*self.layers

        inverse = {}
        for i in range(self.twists):
            inverse[i*self.turns%self.twists] = i

        arms_layers = [[], []]
        # Array where we collect all gerbonara kicad line and arc objects
        for i in range(self.twists):
            start_angle = i*self.sector_angle
            fold_angle = start_angle + self.sweeping_angle
            end_angle = fold_angle + self.sweeping_angle

            # Handle the spiral arm
            points_layer0, arm_length, angle_refs_layer0 = self.shape.compute_spiral(a1=start_angle, a2=fold_angle, fn=circle_segments)
            x0, y0 = points_layer0[0]
            xn, yn = points_layer0[-1]
            if angle_refs_layer0:
                ref_0, ref_n = angle_refs_layer0[0], angle_refs_layer0[-1]
            else:
                ref_0, ref_n = None, None

            if self.approximate_arcs and isinstance(self.shape, CircleShape):
                footprint.arcs.extend(arc_approximate(points_layer0, self.trace_width, self.layer_pair[0], arc_tolerance))
            else:
                footprint.lines.extend(kicad.make_line(*p1, *p2, self.trace_width, self.layer_pair[0]) for p1, p2 in zip(points_layer0, points_layer0[1:]))

            if self.layers > 1:
                # Handle the returning arm on the bottom layer
                points_layer1, _, angle_refs_layer1 = self.shape.compute_spiral(a1=end_angle, a2=fold_angle, fn=circle_segments)
                points_layer1 = points_layer1[::-1]
                if self.approximate_arcs and isinstance(self.shape, CircleShape):
                    footprint.arcs.extend(arc_approximate(points_layer1, self.trace_width, self.layer_pair[1], arc_tolerance))
                else:
                    footprint.lines.extend(kicad.make_line(*p1, *p2, self.trace_width, self.layer_pair[1]) for p1, p2 in zip(points_layer1, points_layer1[1:]))

            else:
                # Add a straight connecting segment connecting the inner point to the outside of the spiral.
                ref = angle_refs_layer0[-1]
                xq, yq = self.shape.project_point(self.shape.outer_radius, fold_angle, r_ref=ref)
                angle_refs_layer1 = [ref, ref]
                points_layer1 = [(xn, yn), (xq, yq)]
                footprint.lines.append(kicad.make_line(xn, yn, xq, yq, self.trace_width, self.layer_pair[1]))
            
            arms_layers[0].append((points_layer0, angle_refs_layer0))
            arms_layers[1].append((points_layer1, angle_refs_layer1))

        for i in range(self.twists):
            start_angle = i*self.sector_angle
            fold_angle = start_angle + self.sweeping_angle
            end_angle = fold_angle + self.sweeping_angle
            
            # Handle inner via ring and process staggering if enabled
            r = self.inner_via_ring_radius
            if self.stagger_inner_vias:
                if i%2 != 0:
                    r -= 2*self.via_offset

            points_layer0, refs_layer0 = arms_layers[0][i]
            points_layer1, refs_layer1 = arms_layers[1][i]

            xv, yv = self.shape.project_point(r, fold_angle, r_ref=refs_layer0[-1])

            footprint.lines.append(kicad.make_line(*points_layer0[-1], xv, yv, self.trace_width, self.layer_pair[0]))
            footprint.lines.append(kicad.make_line(xv, yv, *points_layer1[0], self.trace_width, self.layer_pair[1]))

            footprint.pads.append(kicad.make_via(xv, yv,
                                                 self.via_diameter, self.via_drill, self.clearance,
                                                 self.layer_pair))

            # Handle outer via ring and process staggering if enabled unless we are at the start of the coil, where we will
            # place pads below.
            r = self.outer_via_ring_radius

            if self.stagger_outer_vias:
                if i%2 != 0:
                    r += 2*self.via_offset

            points_layer0, refs_layer0 = arms_layers[0][i]
            points_layer1, refs_layer1 = arms_layers[1][(i - self.turns) % self.twists]
            
            xv, yv = self.shape.project_point(r, start_angle, r_ref=refs_layer0[0])

            footprint.lines.append(kicad.make_line(*points_layer0[0], xv, yv, self.trace_width, self.layer_pair[0]))
            footprint.lines.append(kicad.make_line(*points_layer1[-1], xv, yv, self.trace_width, self.layer_pair[1]))

            if i > 0:
                footprint.pads.append(kicad.make_via(xv, yv,
                                                     self.via_diameter, self.via_drill, self.clearance,
                                                     self.layer_pair))
            
            else:
                # Place the pads on the outer radius
                px, py = self.shape.project_point(self.shape.outer_radius, 0)
                footprint.pads.extend([
                        kicad.make_pad(1, [self.layer_pair[0]], px, py, self.trace_width, self.clearance),
                        kicad.make_pad(2, [self.layer_pair[1]], px, py, self.trace_width, self.clearance)])

        if self.keepout_zone:
            pts = self.shape.offset_exterior(self.keepout_margin)
            footprint.zones.append(Zone(layers=['*.Cu'],
                hatch=Hatch(),
                filled_areas_thickness=False,
                keepout=ZoneKeepout(copperpour_allowed=False),
                polygon=ZonePolygon(pts=[XYCoord(x=x, y=y) for x, y in pts])))

        return footprint




# Copyright 2025 Jan Sebastian Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import subprocess
import webbrowser
import tempfile
import json
import os
import sys
from pathlib import Path
from collections import defaultdict
import warnings
import math

import click
from gerbonara.layers import LayerStack
from gerbonara.cad.kicad.primitives import kicad_mid_to_center_arc

from .geometry import PlanarInductor, divisors, CircleShape, SectorShape, StarShape, SVGShape, RectangleShape, RegularPolygonShape
from .kicad import footprint_to_board
from .svg import make_transparent_svg


def print_valid_twists(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    click.echo(f'Valid twist counts for {value} turns:', file=sys.stderr)
    for d in divisors(value, value):
        click.echo(f'  {d}', file=sys.stderr)

    ctx.exit()


def circle_center_to_tangents(center, a, b):
    """ Given two points on a circle and the center of the circle, calculate the intersection of two tangents at the two points """
    cx, cy = center
    ax, ay = a
    bx, by = b

    dax = ax - cx
    day = ay - cy
    dbx = bx - cx
    dby = by - cy

    v = dax*ax + day*ay
    w = dbx*bx + dby*by
    det = dax*dby - day*dbx

    ix = (v*dby - day*w) / det
    iy = (dax*w - v*dbx) / det

    return ix, iy


@click.group()
@click.option('--turns', type=int, default=5, help='Number of turns')
@click.option('--twists', type=int, default=1, help='Number of twists per revolution. Note that this number must be co-prime to the number of turns. Run with --show-twists to list valid values. (default: 1)')
@click.option('--clockwise/--counter-clockwise', help='Direction of generated top layer spiral. Default: counter-clockwise when wound from the inside.')
@click.option('--single-layer/--two-layer', help='Single-layer mode. This just forces twists to 0.')
@click.option('--show-twists', callback=print_valid_twists, expose_value=False, type=int, is_eager=True, help='Calculate and show valid --twists counts for the given number of turns. Takes the number of turns as a value.')
@click.option('--stagger-inner-vias/--no-stagger-inner-vias', default=False, help='Stagger inner via ring')
@click.option('--stagger-outer-vias/--no-stagger-outer-vias', default=False, help='Stagger outer via ring')
@click.option('--trace-width', type=float, default=None)
@click.option('--clearance', type=float, default=None)
@click.option('--via-diameter', type=float, default=0.6)
@click.option('--via-drill', type=float, default=None)
@click.option('--via-offset', type=float, default=None, help='Radially offset vias from trace endpoints [mm]')
@click.option('--keepout-zone/--no-keepout-zone', default=True, help='Add a keepout are to the footprint (default: yes)')
@click.option('--keepout-margin', type=float, default=5, help='Margin between outside of coil and keepout area (mm, default: 5)')
@click.option('--copper-thickness', type=float, default=0.035, help='Copper thickness for resistance calculation, in mm. Default: 0.035mm ^= 1 Oz')
@click.option('--circle-segments', type=int, default=64, help='When not using arcs, the number of points to use for arc interpolation per 360 degrees.')
@click.option('--arc-tolerance', type=float, default=0.02)
@click.option('--approximate-arcs/--no-approximate-arcs', default=True, help='Use circular arcs to smoothen output shape (default: on)')
@click.option('--format', type=click.Choice(['svg', 'gerber', 'kicad-footprint', 'kicad-pcb', 'json', 'gdsii', 'oasis', 'show']), default='kicad-footprint')
@click.option('--clipboard/--no-clipboard', help='Use clipboard integration (requires wl-clipboard)')
@click.option('--footprint-name', help="Name for the generated footprint. Default: Output file name sans extension.")
@click.option('--cell-name', help="Name for the generated cell when exporting GDSII. Default: Output file name sans extension.")
@click.option('--layer-pair', default='F.Cu,B.Cu', help="Target KiCad layer pair for the generated footprint, comma-separated. Default: F.Cu/B.Cu.")
@click.version_option()
@click.pass_context
def cli(ctx, footprint_name, cell_name, clipboard, single_layer, arc_tolerance, circle_segments, format, **kwargs):
    ctx.ensure_object(dict)
    logger = logging.getLogger('kicoil')
    logger.setLevel(logging.INFO)

    def write(shape, outfile):
        nonlocal footprint_name, clipboard, single_layer, arc_tolerance, circle_segments, format, cell_name
        logger = logging.getLogger('kicoil')

        if single_layer:
            kwargs['layers'] = 1
        else:
            kwargs['layers'] = 2

        try:
            model = PlanarInductor(shape=shape, **kwargs)

            if footprint_name is None and outfile:
                footprint_name = outfile.stem

            footprint = model.render_footprint(footprint_name, arc_tolerance, circle_segments)

        except ValueError as e:
            #raise click.ClickException(*e.args)
            raise

        data = None
        if format == 'kicad-footprint':
            data = footprint.serialize()

        elif format == 'kicad-pcb':
            data = footprint_to_board(footprint).serialize()

        elif format == 'gerber':
            stack = LayerStack()
            footprint.render(stack)

            if not clipboard and outfile and outfile.suffix.lower() != '.zip':
                stack.save_to_directory(outfile)
                return

            else:
                with tempfile.NamedTemporaryFile(delete_on_close=False) as f:
                    f = Path(f.name)
                    stack.save_to_zipfile(f)
                    data = f.read_bytes()
        
        elif format == 'json':
            lines = defaultdict(lambda: [])
            for l in footprint.lines:
                lines[l.layer].append({
                    'x1': l.start.x,
                    'y1': l.start.y,
                    'x2': l.end.x,
                    'y2': l.end.y})
            
            arcs = defaultdict(lambda: [])
            for a in footprint.arcs:
                center, r, direction = kicad_mid_to_center_arc(a.mid, a.start, a.end)
                arcs[a.layer].append({
                    'x1': a.start.x,
                    'y1': a.start.y,
                    'x2': a.end.x,
                    'y2': a.end.y,
                    'cx': center[0],
                    'cy': center[1],
                })

            vias = [{
                    'x': p.at.x,
                    'y': p.at.y,
                    'pad': p.size.x,
                    'drill': p.drill.diameter,
                } for p in footprint.pads if p.number == 'NC']
            
            pads = [{
                    'x': p.at.x,
                    'y': p.at.y,
                    'pad': p.size.x,
                } for p in footprint.pads if p.number != 'NC']
            
            d = {
                'lines': dict(lines),
                'arcs': dict(arcs),
                'vias': vias,
                'pads': pads
            }

            data = json.dumps(d, indent=4)

        elif format in ('gdsii', 'oasis'):
            import gdstk

            DRILL_LAYER = 2
            lib = gdstk.Library()

            if cell_name is None:
                if outfile:
                    cell_name = outfile.stem
                else:
                    cell_name = f'planar_coil'
            cell = lib.new_cell(cell_name)
            
            for line in footprint.lines:
                layer = model.layer_pair.index(line.layer)
                path = gdstk.FlexPath([(line.start.x, line.start.y), (line.end.x, line.end.y)], line.stroke.width, ends=['round'], layer=layer)
                cell.add(path)
            
            for arc in footprint.arcs:
                layer = model.layer_pair.index(arc.layer)
                center, r, _direction = kicad_mid_to_center_arc(arc.mid, arc.start, arc.end)
                proj_x, proj_y = circle_center_to_tangents(center, tuple(arc.start), tuple(arc.end))
                # multiply r with 0.99 to make sure gdstk's interpolation routine catches on since the arc endpoints are calculated exactly
                path = gdstk.FlexPath([(arc.start.x, arc.start.y), (proj_x, proj_y), (arc.end.x, arc.end.y)], arc.stroke.width, bend_radius=r*0.99, ends=['round'], layer=layer)
                cell.add(path)

            for pad in footprint.pads:
                for layer in pad.layers:
                    layer = model.layer_pair.index(layer)
                    layer_obj = gdstk.ellipse(tuple(pad.at), pad.size.x/2, layer=layer)
                    cell.add(layer_obj)
                
                if pad.drill:
                    drill = gdstk.ellipse(tuple(pad.at), pad.drill.diameter/2, layer=DRILL_LAYER)
                    cell.add(drill)

            if clipboard or not outfile:
                raise click.ClickException('outfile is required for GDSII or OASIS export')

            if format == 'gdsii':
                lib.write_gds(outfile)
            else:
                lib.write_oas(outfile)
            return

        elif format in ('svg', 'show'):
            data = str(make_transparent_svg(footprint))

            if format == 'show':
                with tempfile.NamedTemporaryFile('w', suffix='.svg', delete=False) as f:
                    f.write(data)
                    f.flush()
                    webbrowser.open_new_tab(f'file://{f.name}')
                    return

        if clipboard:
            if 'WAYLAND_DISPLAY' in os.environ:
                copy, paste, cliputil = ['wl-copy'], ['wl-paste'], 'xclip'
            else:
                copy, paste, cliputil = ['xclip', '-i', '-sel', 'clipboard'], ['xclip', '-o', '-sel' 'clipboard'], 'wl-clipboard'

            try:
                logger.info(f'Running {copy[0]}.', file=sys.stderr)
                proc = subprocess.Popen(copy, stdin=subprocess.PIPE, text=isinstance(data, str))
                proc.communicate(data)

            except FileNotFoundError:
                raise click.ClickException(f'Error: --clipboard requires the {copy[0]} and {paste[0]} utilities from {cliputil} to be installed.', file=sys.stderr)

        elif not outfile:
            if isinstance(data, str):
                print(data)
            else:
                sys.stdout.buffer.write(data)

        else:
            outfile.write_text(data)
        
    ctx.obj['write'] = write

@cli.command()
@click.option('--outer-diameter', type=float, default=50, help='Outer diameter [mm]')
@click.option('--inner-diameter', type=float, default=25, help='Inner diameter [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def circle(ctx, inner_diameter, outer_diameter, outfile):
    shape = CircleShape(outer_diameter, inner_diameter)
    ctx.obj['write'](shape, outfile)

@cli.command()
@click.option('--width', type=float, default=50, help='Base width [mm]')
@click.option('--height', type=float, default=40, help='Shape height [mm]')
@click.option('--offset', type=float, default=10, help='Offset of each corner at the shorter edge compared to the longer edge [mm]')
@click.option('--annular-width', type=float, default=10, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def trapezoid(ctx, outfile, **kwargs):
    shape = TrapezoidShape(**kwargs)
    ctx.obj['write'](shape, outfile)


@cli.command()
@click.option('--width', type=float, default=50, help='Width [mm]')
@click.option('--height', type=float, default=40, help='Height [mm]')
@click.option('--annular-width', type=float, default=10, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def rectangle(ctx, outfile, **kwargs):
    shape = RectangleShape(**kwargs)
    ctx.obj['write'](shape, outfile)


@cli.command()
@click.option('--diameter', type=float, default=50, help='Width [mm]')
@click.option('-n', '--corners', type=int, default=8, help='Number of corners')
@click.option('--annular-width', type=float, default=10, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def regular_polygon(ctx, outfile, **kwargs):
    shape = RegularPolygonShape(**kwargs)
    ctx.obj['write'](shape, outfile)


@cli.command()
@click.option('--inner-diameter', type=float, default=25, help='Inner diameter [mm]')
@click.option('--outer-diameter', type=float, default=50, help='Outer diameter [mm]')
@click.option('--angle', type=float, default=45, help='Sector angle [deg]')
@click.option('--arc-tolerance', type=float, default=0.05, help='Tolerance for splitting arc into straight segments [mm] (default: 0.05 mm)')
@click.option('--annular-width', type=float, default=5, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def sector(ctx, outfile, angle, **kwargs):
    angle = math.radians(angle)
    shape = SectorShape(angle=angle, **kwargs)
    ctx.obj['write'](shape, outfile)


@cli.command()
@click.option('--inner-diameter', type=float, default=25, help='Inner diameter [mm]')
@click.option('--outer-diameter', type=float, default=50, help='Outer diameter [mm]')
@click.option('--points', type=int, default=5, help='Number of points')
@click.option('--arc-tolerance', type=float, default=0.05, help='Tolerance for splitting arc into straight segments [mm] (default: 0.05 mm)')
@click.option('--annular-width', type=float, default=5, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def star(ctx, outfile, **kwargs):
    shape = StarShape(**kwargs)
    ctx.obj['write'](shape, outfile)


@cli.command()
@click.option('--arc-tolerance', type=float, default=0.05, help='Tolerance for splitting arc into straight segments [mm] (default: 0.05 mm)')
@click.option('--annular-width', type=float, default=5, help='Width of the trace area on the outside of the shape [mm]')
@click.argument('svg_file', required=False, type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument('outfile', required=False, type=click.Path(writable=True, dir_okay=False, path_type=Path))
@click.pass_context
def svg(ctx, svg_file, outfile, **kwargs):
    shape = SVGShape(svg_file, **kwargs)
    ctx.obj['write'](shape, outfile)



# Copyright 2025 Jan Sebastian Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file was created using claude.ai because hand-writing GUI code sucks. The rest of kicoil is written by hand.
#

""" GUI for generating KiCad footprints using kicoil. """

import tkinter as tk
from tkinter import ttk, filedialog, messagebox, scrolledtext
import sys
import logging
import warnings
import traceback
from pathlib import Path
from contextlib import contextmanager
from io import BytesIO

from .geometry import PlanarInductor, divisors, CircleShape
from .svg import make_transparent_svg

try:
    # for rendering gerbonara's svg output to PNG
    import cairosvg
    # for scaling the rendered PNGs to the UI's resolution since tkinter is very limited there
    from PIL import Image, ImageTk
    HAS_PREVIEW = True
except ImportError:
    HAS_PREVIEW = False


class TextWidgetHandler(logging.Handler):
    """Custom logging handler that writes to a tkinter Text widget"""
    def __init__(self, text_widget):
        super().__init__()
        self.text_widget = text_widget
        self.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
        
    def emit(self, record):
        msg = self.format(record)

        # Determine tag based on log level
        if record.levelno >= logging.ERROR:
            tag = 'error'
        elif record.levelno >= logging.WARNING:
            tag = 'warning'
        else:
            tag = 'info'

        # Temporarily enable widget to insert text
        self.text_widget['state'] = 'normal'
        self.text_widget.insert(tk.END, msg + '\n', tag)
        self.text_widget.see(tk.END)
        self.text_widget['state'] = 'disabled'
        self.text_widget.update_idletasks()


# https://stackoverflow.com/questions/27820178/how-to-add-placeholder-to-an-entry-in-tkinter
class EntryWithPlaceholder(tk.Entry):
    def __init__(self, master=None, placeholder="empty", color='grey', *args, **kwargs):
        # Initialize parent Entry with all provided arguments
        super().__init__(master, *args, **kwargs)

        self.placeholder = placeholder
        self.placeholder_color = color
        self.showing_placeholder = False

        # Get the default foreground color from the created widget
        self.default_fg_color = self['fg']

        self.bind("<FocusIn>", self.foc_in)
        self.bind("<FocusOut>", self.foc_out)

        # Check initial state and show placeholder if empty
        # Use a small delay to ensure the widget is fully initialized
        self.after(1, self._check_initial_state)

    def _check_initial_state(self):
        if not super().get():
            self.put_placeholder()

    def put_placeholder(self):
        self.showing_placeholder = True
        self['fg'] = self.placeholder_color
        super().delete(0, 'end')
        super().insert(0, self.placeholder)

    def update_placeholder(self, new_placeholder):
        self.placeholder = new_placeholder
        # If currently showing placeholder, update the display
        if self.showing_placeholder:
            super().delete(0, 'end')
            super().insert(0, self.placeholder)

    def foc_in(self, *args):
        if self.showing_placeholder:
            self.showing_placeholder = False
            self['fg'] = self.default_fg_color
            super().delete(0, 'end')

    def foc_out(self, *args):
        if not super().get():
            self.put_placeholder()
        else:
            self.showing_placeholder = False

    def get(self):
        if self.showing_placeholder:
            return ''
        return super().get()


class KiCoilGUI:
    def __init__(self, root, kicad_inst=None):
        self.kicad_inst = kicad_inst
        self.root = root
        self.root.title("KiCoil - Planar Inductor Generator")
        self.root.geometry("1000x650")

        style = ttk.Style()
        style.theme_use('clam')

        main_container = ttk.Frame(root)
        main_container.grid(row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)
        main_container.columnconfigure(0, weight=0)  # Left panel doesn't grow horizontally
        main_container.columnconfigure(1, weight=1)  # Preview panel grows to fill space
        main_container.rowconfigure(0, weight=1)

        main_frame = ttk.Frame(main_container, padding="10")
        main_frame.grid(row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))

        title_frame = ttk.Frame(main_frame)
        title_frame.grid(row=0, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=(0, 10))
        ttk.Label(title_frame, text="Planar Inductor Generator",
                 font=('Helvetica', 16, 'bold')).pack(side=tk.LEFT)

        self.preview_visible = tk.BooleanVar(value=True)
        self.preview_button = ttk.Button(title_frame, text="Hide Preview",
                  command=self.toggle_preview, width=15)
        self.preview_button.pack(side=tk.RIGHT)

        self.notebook = ttk.Notebook(main_frame)
        self.notebook.grid(row=1, column=0, columnspan=3, sticky=(tk.W, tk.E), pady=5)

        geometry_frame = ttk.Frame(self.notebook, padding="10")
        self.notebook.add(geometry_frame, text="Geometry")
        self.create_geometry_params(geometry_frame)

        traces_frame = ttk.Frame(self.notebook, padding="10")
        self.notebook.add(traces_frame, text="Traces")
        self.create_trace_params(traces_frame)

        via_frame = ttk.Frame(self.notebook, padding="10")
        self.notebook.add(via_frame, text="Vias")
        self.create_via_params(via_frame)

        output_frame = ttk.Frame(self.notebook, padding="10")
        self.notebook.add(output_frame, text="Output")
        self.create_output_params(output_frame)
        
        button_frame = ttk.Frame(main_frame)
        button_frame.grid(row=2, column=0, columnspan=3, pady=10)
        
        ttk.Button(button_frame, text="Show Valid Twists", 
                  command=self.show_valid_twists, width=20).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Save Footprint File", 
                  command=self.save_footprint_file, width=20).pack(side=tk.LEFT, padx=5)
        #if self.kicad_inst: FIXME
        #    ttk.Button(button_frame, text="Update Footprint on Board", 
        #              command=self.update_board_footprint, width=20).pack(side=tk.LEFT, padx=5)
        
        status_label = ttk.Label(main_frame, text="Output:", font=('Helvetica', 10, 'bold'))
        status_label.grid(row=3, column=0, sticky=tk.W, pady=(10, 0))
        
        self.output_text = scrolledtext.ScrolledText(main_frame, height=8, width=80, state='disabled')
        self.output_text.grid(row=4, column=0, columnspan=3, sticky=(tk.W, tk.E, tk.N, tk.S), pady=5)
        self.output_text.tag_config('error', foreground='red')
        self.output_text.tag_config('warning', foreground='orange')
        self.output_text.tag_config('info', foreground='black')

        main_frame.columnconfigure(0, weight=1)
        main_frame.rowconfigure(1, weight=0)  # Notebook doesn't grow vertically
        main_frame.rowconfigure(4, weight=1)  # Output area takes all vertical space

        self.preview_frame = ttk.LabelFrame(main_container, text="Preview", padding=10)

        if HAS_PREVIEW:
            # Create canvas for image display
            self.preview_canvas = tk.Canvas(self.preview_frame, bg='white')
            self.preview_canvas.pack(fill=tk.BOTH, expand=True)
            self.preview_image = None  # Store reference to prevent garbage collection
            self.preview_raw_image = None  # Store unscaled image for rescaling

            # Bind canvas resize event to update preview
            self.preview_canvas.bind('<Configure>', self._on_preview_resize)
        
        else:
            info_text = "Preview not available\n\nInstall dependencies:\npip install cairosvg pillow"
            self.preview_label = ttk.Label(self.preview_frame, text=info_text,
                                           justify=tk.CENTER, anchor=tk.CENTER)
            self.preview_label.pack(fill=tk.BOTH, expand=True)

        self.preview_frame.grid(row=0, column=1, sticky=(tk.N, tk.S, tk.E, tk.W), padx=(10, 10), pady=10)
                
        self.current_model = None
        self._validation_after_id = None
        self.setup_logging()
        self.setup_traces()
        self.root.after(100, self.validate_parameters)

    def _on_preview_resize(self, event):
        # Debounce resize events - only update after resize is complete
        if hasattr(self, '_resize_after_id') and self._resize_after_id is not None:
            self.root.after_cancel(self._resize_after_id)
        self._resize_after_id = self.root.after(100, self._rescale_preview)

    def _rescale_preview(self):
        canvas_width = self.preview_canvas.winfo_width()
        canvas_height = self.preview_canvas.winfo_height()
        if canvas_width <= 1 or canvas_height <= 1 or self.preview_raw_image is None:
            self.preview_canvas.delete("all")
            return

        # Calculate scaling to fit within canvas
        image = self.preview_raw_image.copy()
        img_ratio = image.width / image.height
        canvas_ratio = canvas_width / canvas_height

        if img_ratio > canvas_ratio:
            # Image is wider than canvas
            new_width = int(canvas_width * 0.95)
            new_height = int(new_width / img_ratio)
        else:
            # Image is taller than canvas
            new_height = int(canvas_height * 0.95)
            new_width = int(new_height * img_ratio)

        image = image.resize((new_width, new_height), Image.Resampling.LANCZOS)

        self.preview_image = ImageTk.PhotoImage(image)
        self.preview_canvas.create_image(
            canvas_width // 2, canvas_height // 2,
            image=self.preview_image, anchor=tk.CENTER
        )

    def toggle_preview(self):
        if self.preview_visible.get():
            self.preview_frame.grid_forget()
            self.preview_visible.set(False)
            self.preview_button.config(text="Show Preview")

            current_width = self.root.winfo_width()
            current_height = self.root.winfo_height()
            new_width = max(680, current_width - 400)
            self.root.geometry(f"{new_width}x{current_height}")

        else:
            self.preview_frame.grid(row=0, column=1, sticky=(tk.N, tk.S, tk.E, tk.W), padx=(10, 0), pady=10)
            self.preview_visible.set(True)
            self.preview_button.config(text="Hide Preview")

            current_width = self.root.winfo_width()
            current_height = self.root.winfo_height()
            new_width = current_width + 400  # Add ~400px for preview
            self.root.geometry(f"{new_width}x{current_height}")

    def create_geometry_params(self, parent):
        row = 0

        # Turns
        ttk.Label(parent, text="Number of Turns:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.turns_var = tk.IntVar(value=7)
        ttk.Spinbox(parent, from_=1, to=100, textvariable=self.turns_var,
                   width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Number of spiral turns",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1

        # Twists
        ttk.Label(parent, text="Twists per Revolution:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.twists_var = tk.IntVar(value=4)
        ttk.Spinbox(parent, from_=0, to=50, textvariable=self.twists_var,
                   width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Must be co-prime to turns",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1

        # Outer Diameter
        ttk.Label(parent, text="Outer Diameter (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.outer_dia_var = tk.DoubleVar(value=50.0)
        ttk.Spinbox(parent, from_=1, to=500, increment=0.5,
                   textvariable=self.outer_dia_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Outside diameter of coil",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1

        # Inner Diameter
        ttk.Label(parent, text="Inner Diameter (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.inner_dia_var = tk.DoubleVar(value=25.0)
        ttk.Spinbox(parent, from_=0, to=500, increment=0.5,
                   textvariable=self.inner_dia_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Inside diameter of coil",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1

        # Layer Mode
        ttk.Label(parent, text="Layer Mode:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.layer_mode_var = tk.IntVar(value=2)
        layer_frame = ttk.Frame(parent)
        layer_frame.grid(row=row, column=1, columnspan=2, sticky=tk.W, pady=5)
        ttk.Radiobutton(layer_frame, text="Two Layer", variable=self.layer_mode_var,
                       value=2).pack(side=tk.LEFT, padx=(0, 10))
        ttk.Radiobutton(layer_frame, text="Single Layer", variable=self.layer_mode_var,
                       value=1).pack(side=tk.LEFT)
        row += 1

        # Direction
        ttk.Label(parent, text="Winding Direction:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.direction_var = tk.StringVar(value="counter-clockwise")
        dir_frame = ttk.Frame(parent)
        dir_frame.grid(row=row, column=1, columnspan=2, sticky=tk.W, pady=5)
        ttk.Radiobutton(dir_frame, text="Counter-Clockwise", variable=self.direction_var,
                       value="counter-clockwise").pack(side=tk.LEFT, padx=(0, 10))
        ttk.Radiobutton(dir_frame, text="Clockwise", variable=self.direction_var,
                       value="clockwise").pack(side=tk.LEFT)
        row += 1

    def create_trace_params(self, parent):
        row = 0

        # Trace Width
        ttk.Label(parent, text="Trace Width (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.trace_width_entry = EntryWithPlaceholder(parent, placeholder="automatic", width=15)
        self.trace_width_entry.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Clearance
        ttk.Label(parent, text="Clearance (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.clearance_entry = EntryWithPlaceholder(parent, placeholder="automatic", width=15)
        self.clearance_entry.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Copper Thickness (in µm)
        ttk.Label(parent, text="Copper Thickness (µm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.copper_thickness_var = tk.DoubleVar(value=35.0)  # 35µm = 0.035mm = 1 Oz
        ttk.Spinbox(parent, from_=1, to=1000, increment=1, format="%.1f",
                   textvariable=self.copper_thickness_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="35µm = 1 Oz copper",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1
        
    def create_via_params(self, parent):
        """Create via parameter controls"""
        row = 0
        
        # Via Diameter
        ttk.Label(parent, text="Via Diameter (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.via_diameter_var = tk.DoubleVar(value=0.6)
        ttk.Spinbox(parent, from_=0.1, to=5.0, increment=0.1, format="%.2f",
                   textvariable=self.via_diameter_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1
        
        # Via Drill
        ttk.Label(parent, text="Via Drill (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.via_drill_entry = EntryWithPlaceholder(parent, placeholder="automatic", width=15)
        self.via_drill_entry.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Via Offset
        ttk.Label(parent, text="Via Offset (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.via_offset_entry = EntryWithPlaceholder(parent, placeholder="automatic", width=15)
        self.via_offset_entry.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1
        
        # Stagger Inner Vias
        self.stagger_inner_var = tk.BooleanVar(value=False)
        ttk.Checkbutton(parent, text="Stagger inner via ring", 
                       variable=self.stagger_inner_var).grid(row=row, column=0, columnspan=2, sticky=tk.W, pady=5)
        row += 1
        
        # Stagger Outer Vias
        self.stagger_outer_var = tk.BooleanVar(value=False)
        ttk.Checkbutton(parent, text="Stagger outer via ring", 
                       variable=self.stagger_outer_var).grid(row=row, column=0, columnspan=2, sticky=tk.W, pady=5)
        row += 1
        
    def create_output_params(self, parent):
        row = 0

        # Footprint Name
        ttk.Label(parent, text="Footprint Name:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.footprint_name_entry = EntryWithPlaceholder(parent, placeholder="automatic")
        self.footprint_name_entry.grid(row=row, column=1, columnspan=2, sticky=(tk.W, tk.E), pady=5)
        row += 1

        # KiCad layer names
        copper_layers = ['F.Cu', 'B.Cu'] + [f'In{i}.Cu' for i in range(1, 31)]

        # Top Layer
        ttk.Label(parent, text="Top Layer:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.top_layer_var = tk.StringVar(value="F.Cu")
        top_layer_combo = ttk.Combobox(parent, textvariable=self.top_layer_var,
                                        values=copper_layers, state='readonly', width=23)
        top_layer_combo.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Bottom Layer
        ttk.Label(parent, text="Bottom Layer:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.bottom_layer_var = tk.StringVar(value="B.Cu")
        bottom_layer_combo = ttk.Combobox(parent, textvariable=self.bottom_layer_var,
                                           values=copper_layers, state='readonly', width=23)
        bottom_layer_combo.grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Circle Segments
        ttk.Label(parent, text="Circle Segments:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.circle_segments_var = tk.IntVar(value=64)
        ttk.Spinbox(parent, from_=8, to=360, textvariable=self.circle_segments_var,
                   width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Points per 360° for arc interpolation",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1

        # Arc Tolerance
        ttk.Label(parent, text="Arc Tolerance (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.arc_tolerance_var = tk.DoubleVar(value=0.02)
        ttk.Spinbox(parent, from_=0.001, to=1.0, increment=0.001, format="%.3f",
                   textvariable=self.arc_tolerance_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Keepout Zone
        ttk.Label(parent, text="Keepout Zone:").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.keepout_var = tk.BooleanVar(value=True)
        ttk.Checkbutton(parent, text="Add keepout area",
                       variable=self.keepout_var).grid(row=row, column=1, sticky=tk.W, pady=5)
        row += 1

        # Keepout Margin
        ttk.Label(parent, text="Keepout Margin (mm):").grid(row=row, column=0, sticky=tk.W, pady=5)
        self.keepout_margin_var = tk.DoubleVar(value=5.0)
        ttk.Spinbox(parent, from_=0, to=50, increment=0.5,
                   textvariable=self.keepout_margin_var, width=15).grid(row=row, column=1, sticky=tk.W, pady=5)
        ttk.Label(parent, text="Margin around coil",
                 foreground='gray').grid(row=row, column=2, sticky=tk.W, padx=(10, 0))
        row += 1
        
    def get_parameters(self):
        shape_params = {
            'outer_diameter'     : self.outer_dia_var.get(),
            'inner_diameter'     : self.inner_dia_var.get(),
        }

        params = {
            'shape'              : CircleShape(**shape_params),
            'turns'              : self.turns_var.get(),
            'layers'             : self.layer_mode_var.get(),
            'twists'             : self.twists_var.get(),
            'clockwise'          : (self.direction_var.get() == "clockwise"),
            'copper_thickness'   : self.copper_thickness_var.get() / 1000.0,  # µm -> mm
            'keepout_zone'       : self.keepout_var.get(),
            'keepout_margin'     : self.keepout_margin_var.get(),
            'via_diameter'       : self.via_diameter_var.get(),
            'stagger_inner_vias' : self.stagger_inner_var.get(),
            'stagger_outer_vias' : self.stagger_outer_var.get(),
            'layer_pair'         : f"{self.top_layer_var.get()},{self.bottom_layer_var.get()}",
        }

        if (trace_width_value := self.trace_width_entry.get()):
            params['trace_width'] = float(trace_width_value)

        if (clearance_value := self.clearance_entry.get()):
            params['clearance'] = float(clearance_value)

        if (via_drill_value := self.via_drill_entry.get()):
            params['via_drill'] = float(via_drill_value)

        if (via_offset_value := self.via_offset_entry.get()):
            params['via_offset'] = float(via_offset_value)

        return params
    
    def setup_logging(self):
        """Set up logging handler to capture kicoil logger output for display in the output text widget"""
        self.kicoil_logger = logging.getLogger('kicoil')
        self.kicoil_logger.setLevel(logging.INFO)
        self.kicoil_logger.handlers.clear()
        self.log_handler = TextWidgetHandler(self.output_text)
        self.kicoil_logger.addHandler(self.log_handler)
    
    @contextmanager
    def capture_warnings(self):
        """Context manager to capture kicoil's warnings to the output text widget"""
        def show_warning(message, category, filename, lineno, file=None, line=None):
            self.output_text['state'] = 'normal'
            self.output_text.insert(tk.END, f'{message}\n', 'warning')
            self.output_text.see(tk.END)
            self.output_text['state'] = 'disabled'
            self.output_text.update_idletasks()

        old_showwarning, warnings.showwarning = warnings.showwarning, show_warning
        try:
            yield
        finally:
            warnings.showwarning = old_showwarning
    
    def setup_traces(self):
        for var in [
                    self.turns_var,
                    self.outer_dia_var,
                    self.inner_dia_var,
                    self.layer_mode_var,
                    self.direction_var,
                    self.twists_var,
                    self.copper_thickness_var,
                    self.keepout_var,
                    self.keepout_margin_var,
                    self.via_diameter_var,
                    self.stagger_inner_var,
                    self.stagger_outer_var,
                    self.top_layer_var,
                    self.bottom_layer_var]:
            var.trace_add('write', self._on_parameter_change)

        for entry in [self.trace_width_entry, self.clearance_entry,
                      self.via_drill_entry, self.via_offset_entry,
                      self.footprint_name_entry]:
            entry.bind('<KeyRelease>', lambda e: self._on_parameter_change())
    
    def _on_parameter_change(self, *args):
        # Schedule validation to avoid too many rapid calls
        if self._validation_after_id is not None:
            self.root.after_cancel(self._validation_after_id)
        self._validation_after_id = self.root.after(200, self.validate_parameters)
    
    def validate_parameters(self):
        """Validate parameters by creating PlanarInductor instance"""
        try:
            self.output_text['state'] = 'normal'
            self.output_text.delete('1.0', tk.END)

            with self.capture_warnings():
                self.current_model = PlanarInductor(**self.get_parameters())

            # If we got here, parameters are valid
            self.output_text.insert(tk.END, "Parameters valid\n", 'info')
            self.update_placeholders()
            self.update_preview()
            return True

        except ValueError as e:
            self.output_text.insert(tk.END, f"ERROR: {e}\n", 'error')
            self.output_text.see(tk.END)

            self.current_model = None
            self.update_placeholders()
            return False
            
        except Exception as e:
            tb = traceback.format_exc()
            self.output_text.insert(tk.END, f"Unexpected error:\n{tb}\n", 'error')
            self.output_text.see(tk.END)

            print(tb, file=sys.stderr)

            self.current_model = None
            self.update_placeholders()
            return True

        finally:
            self.output_text['state'] = 'disabled'

    def update_placeholders(self):
        if self.current_model is None:
            self.trace_width_entry.update_placeholder("automatic")
            self.clearance_entry.update_placeholder("automatic")
            self.via_drill_entry.update_placeholder("automatic")
            self.via_offset_entry.update_placeholder("automatic")
            self.footprint_name_entry.update_placeholder("automatic")
        
        else:
            if not self.trace_width_entry.get():
                self.trace_width_entry.update_placeholder(f"auto: {self.current_model.trace_width:.2f}")

            if not self.clearance_entry.get():
                self.clearance_entry.update_placeholder(f"auto: {self.current_model.clearance:.2f}")

            if not self.via_drill_entry.get():
                self.via_drill_entry.update_placeholder(f"auto: {self.current_model.via_drill:.2f}")

            if not self.via_offset_entry.get():
                self.via_offset_entry.update_placeholder(f"auto: {self.current_model.via_offset:.2f}")

            if not self.footprint_name_entry.get():
                self.footprint_name_entry.update_placeholder(self.current_model.default_footprint_name)

    def update_preview(self):
        if not HAS_PREVIEW:
            return

        arc_tolerance = self.arc_tolerance_var.get()
        circle_segments = self.circle_segments_var.get()

        footprint = self.current_model.render_footprint(None, arc_tolerance, circle_segments)
        svg_tag = make_transparent_svg(footprint)
        viewbox = svg_tag.attrs.get('viewBox', '0 0 800 800')
        _, _, svg_width, svg_height = map(float, viewbox.split())
        min_dimension = 800
        scale = max(min_dimension / svg_width, min_dimension / svg_height)
        output_width = int(svg_width * scale)
        output_height = int(svg_height * scale)
        svg_tag.attrs['width'] = f'{output_width}px'
        svg_tag.attrs['height'] = f'{output_height}px'

        png_data = cairosvg.svg2png(bytestring=str(svg_tag).encode('utf-8'))
        self.preview_raw_image = Image.open(BytesIO(png_data))
        self._rescale_preview()

    
    def show_valid_twists(self):
        turns = self.turns_var.get()
        valid_twists = list(divisors(turns, turns))

        self.output_text['state'] = 'normal'
        self.output_text.delete('1.0', tk.END)
        self.output_text.insert('1.0', f'Valid twist counts for {turns} turns:\n')
        for d in valid_twists:
            self.output_text.insert(tk.END, f'  {d}\n')
        self.output_text['state'] = 'disabled'
    
    def update_board_footprint(self):
        if not self.validate_parameters():
            messagebox.showerror("Error", "Cannot generate model. Please check the output for warnings or errors.")
            return

        from kipy.board_types import FootprintInstance, Footprint, Pad, BoardArc, BoardSegment, FootprintAttributes,\
                PadStack, PadStackLayer, PadStackType, PadStackShape, DrillProperties, PadType
        from kipy.geometry import Vector2
        from kipy.common_types import GraphicAttributes, StrokeAttributes
        from kipy.util import from_mm
        from kipy.util.board_layer import CANONICAL_LAYER_NAMES

        from gerbonara.cad.kicad.footprints import Atom

        board = self.kicad_inst.get_board()
        selected = [item for item in board.get_selection() if isinstance(item, FootprintInstance)]

        if not selected:
            messagebox.showerror("Error", "No footprint selected. Select one footprint to replace in KiCad's PCB editor.")
            return
        elif len(selected) > 1:
            messagebox.showerror("Error", "More than one footprint selected. Select only the footprint you want to replace.")
            return

        selected_footprint, = selected

        footprint_name = self.footprint_name_entry.get() or None
        arc_tolerance = self.arc_tolerance_var.get()
        circle_segments = self.circle_segments_var.get()

        self.output_text['state'] = 'normal'
        self.output_text.insert(tk.END, "Rendering footprint...\n", 'info')
        self.output_text.see(tk.END)

        model = self.current_model.render_footprint(footprint_name, arc_tolerance, circle_segments)
        selected_footprint.attributes.exclude_from_bill_of_materials = True
        layer_map = {v: k for k, v in CANONICAL_LAYER_NAMES.items()}
        items = []

        for line in model.lines:
            seg = BoardSegment()
            seg.start = Vector2.from_xy(from_mm(line.start.x), from_mm(line.start.y))
            seg.end = Vector2.from_xy(from_mm(line.end.x), from_mm(line.end.y))
            seg.attributes.stroke.width = from_mm(line.stroke.width)
            seg.layer = layer_map[line.layer]
            selected_footprint.definition.add_item(seg)
            items.append(seg)

        for ref in model.arcs:
            arc = BoardArc()
            arc.start = Vector2.from_xy(from_mm(ref.start.x), from_mm(ref.start.y))
            arc.mid = Vector2.from_xy(from_mm(ref.mid.x), from_mm(ref.mid.y))
            arc.end = Vector2.from_xy(from_mm(ref.end.x), from_mm(ref.end.y))
            arc.attributes.stroke.width = from_mm(ref.stroke.width)
            arc.layer = layer_map[ref.layer]
            selected_footprint.definition.add_item(arc)
            items.append(arc)

        for ref in model.pads:
            pad = Pad() 
            pad.number = ref.number
            pad.position = Vector2.from_xy(from_mm(ref.at.x), from_mm(ref.at.y))
            pad.type = PadType.PT_SMD if ref.type == Atom.smd else PadType.PT_PTH
            pad.padstack.type = PadStackType.PST_NORMAL
            pad.padstack.layers = [layer_map[name] for name in ref.layers]
            layer = pad.padstack.copper_layers[0]
            layer.shape = PadStackShape.PSS_CIRCLE
            layer.size = Vector2.from_xy(from_mm(ref.size.x), from_mm(ref.size.y))
            layer.layer = layer_map[ref.layers[0]] # ? duplicate
            if ref.drill:
                pad.padstack.drill.diameter = Vector2.from_xy(from_mm(ref.drill.diameter), from_mm(ref.drill.diameter))
            selected_footprint.definition.add_item(pad)
            items.append(pad)

        commit = board.begin_commit()
        board.create_items(items)
        board.update_items([selected_footprint])
        board.push_commit(commit, 'Updated planar coil footprint')
        self.output_text.insert(tk.END, "Done.", 'info')
        self.output_text['state'] = 'disabled'
        self.output_text.see(tk.END)

    def save_footprint_file(self):
        if not self.validate_parameters():
            messagebox.showerror("Error", "Cannot generate model. Please check the output for warnings or errors.")
            return
        
        try:
            footprint_name = self.footprint_name_entry.get() or None
            arc_tolerance = self.arc_tolerance_var.get()
            circle_segments = self.circle_segments_var.get()

            self.output_text['state'] = 'normal'
            self.output_text.insert(tk.END, "Rendering footprint...\n\n", 'info')

            footprint = self.current_model.render_footprint(footprint_name, arc_tolerance, circle_segments)
            default_name = footprint_name or self.current_model.default_footprint_name
            output_file = filedialog.asksaveasfilename(
                title="Save KiCad Footprint",
                defaultextension=".kicad_mod",
                initialfile=f"{default_name}.kicad_mod",
                filetypes=[("KiCad Footprint", "*.kicad_mod"), ("All files", "*.*")]
            )

            if not output_file:
                self.output_text.insert(tk.END, "\nSave cancelled.\n", 'info')
                return

            Path(output_file).write_text(footprint.serialize())
            self.output_text.insert(tk.END, f"\nSuccess! Footprint saved to:\n  {output_file}\n", 'info')

        except Exception as e:
            tb = traceback.format_exc()
            self.output_text['state'] = 'normal'
            self.output_text.insert(tk.END, f"\nError generating footprint:\n{tb}\n", 'error')
            self.output_text.see(tk.END)

            print(tb, file=sys.stderr)

            messagebox.showerror("Error", f"Error generating footprint: {e}")

        finally:
            self.output_text['state'] = 'disabled'


def main(kicad_inst=None):
    from kipy import KiCad
    from kipy.errors import ConnectionError
    kicad_inst = KiCad()
    root = tk.Tk()
    app = KiCoilGUI(root, kicad_inst)
    root.mainloop()




# Copyright 2025 Jan Sebastian Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import textwrap

from gerbonara.utils import Tag
from gerbonara.layers import LayerStack

def make_transparent_svg(footprint):
    stack = LayerStack()
    footprint.render(stack)
    root_tag = stack.to_svg(margin=5, colors={
           'top copper': '#eb5c5c',
        'bottom copper': '#5c5ceb',
            'drill pth': '#52cc52',
        })

    root_tag.children[0].attrs['opacity'] = '0.7'
    root_tag.children[1].attrs['opacity'] = '0.7'
    root_tag.children[1].attrs['style'] = 'mix-blend-mode: multiply'
    return root_tag

